@@ -15,7 +15,7 @@
 #include <sys/types.h>
 #include <libcli.h>
 
-#define VERSION        "2.3.2"
+#define VERSION        "2.4.1"
 
 // Limits
 #define MAXTUNNEL	500		// could be up to 65535
@@ -105,7 +105,7 @@
 #define ACCT_TIME	3000				// 5 minute accounting interval
 #define ACCT_SHUT_TIME	600				// 1 minute for counters of shutdown sessions
 #define L2TPPORT	1701				// L2TP port
-#define RADPORT		1645				// old radius port...
+#define RADPORT		1812				// radius port
 #define DAEPORT		3799				// DAE port
 #define PKTARP		0x0806				// ARP packet type
 #define PKTIP		0x0800				// IPv4 packet type
@@ -306,7 +306,7 @@ typedef struct
 	uint32_t pin, pout;		// packet counts
 	uint32_t cin, cout;		// byte counts
 	uint32_t cin_wrap, cout_wrap;	// byte counter wrap count (RADIUS accounting giagawords)
-	uint32_t cin_delta, cout_delta;	// byte count changes (for dump_session())
+	uint64_t cin_delta, cout_delta;	// byte count changes (for dump_session())
 	uint16_t throttle_in;		// upstream throttle rate (kbps)
 	uint16_t throttle_out;		// downstream throttle rate
 	uint8_t filter_in;		// input filter index (to ip_filters[N-1]; 0 if none)
