@@ -31,7 +31,8 @@ char *fmtaddr(in_addr_t addr, int n)
 
 char *fmtMacAddr(uint8_t *pMacAddr)
 {
-	static char strMAC[2*ETH_ALEN];
+	// FF:FF:FF:FF:FF:FF (len(18)= 2*6 + 5 (:) + 1 (zero terminal))
+	static char strMAC[2*ETH_ALEN+6];
 
 	sprintf(strMAC, "%02X:%02X:%02X:%02X:%02X:%02X",
 			pMacAddr[0], pMacAddr[1], pMacAddr[2],
